# Playwright 抓取网页


&lt;!--more--&gt;

## 启动 docker 容器

``` shell
# 前台运行,推出后删除容器
docker run -p 3000:3000 --rm --init -it --workdir /home/pwuser --user pwuser mcr.microsoft.com/playwright:v1.51.0-noble /bin/sh -c &#34;npx -y playwright@1.51.0 run-server --port 3000 --host 0.0.0.0&#34;

# 后台运行
docker run --add-host=hostmachine:host-gateway -p 3000:3000 -d --init -it --workdir /home/pwuser --user pwuser mcr.microsoft.com/playwright:v1.52.0-noble /bin/sh -c &#34;npx -y playwright@1.52.0 run-server --port 3000 --host 0.0.0.0&#34;
```

{{&lt; admonition type=tip title=&#34;命令详解&#34; open=true &gt;}}
参数说明
- docker run: 启动一个新的容器。
- -p 3000:3000: 将主机的端口 3000 映射到容器内的端口 3000。这允许外部网络通过主机的 IP 地址和端口 3000 访问运行在容器内的服务。
- --rm: 当容器退出时自动删除该容器。这有助于避免未使用的容器占用系统资源。
- --init: 在容器中使用一个小的初始化进程作为 PID 1，以确保正确处理信号并防止僵尸进程出现。
- -it:
- -i (interactive) 保持 STDIN 开放，即使没有附加也保持打开状态。
- -t (tty) 分配一个伪 TTY（终端），这对于交互式命令非常有用。
- --workdir /home/pwuser: 设置容器内的工作目录为 /home/pwuser。所有后续命令都会在这个目录下执行。
- --user pwuser: 指定容器内运行进程的用户为 pwuser 而不是默认的 root 用户。这是为了提高安全性，尤其是在处理不受信任的内容时。
- mcr.microsoft.com/playwright:v1.51.0-noble: 使用的 Docker 镜像的名字和标签。这里使用的是 Microsoft 提供的 Playwright 版本 1.51.0 的镜像。
- /bin/sh -c &#34;npx -y playwright@1.51.0 run-server --port 3000 --host 0.0.0.0&#34;:


此命令的作用是：
- 启动一个基于 mcr.microsoft.com/playwright:v1.51.0-noble 镜像的 Docker 容器。
- 将主机的 3000 端口映射到容器内的 3000 端口。
- 使用 pwuser 用户而非 root 用户来运行容器内的进程。
- 设置容器的工作目录为 /home/pwuser。
- 使用 /bin/sh shell 执行命令：使用 npx 工具以非交互方式安装并运行 Playwright 服务器，监听在 3000 端口且可被外部访问（通过设置 --host 0.0.0.0）。
{{&lt; /admonition &gt;}}




## 在 python 中使用

### 测试用例

``` python {data-open=true,title=&#34;测试用例&#34;}
import unittest
from playwright.async_api import async_playwright

class TestPlaywright(unittest.IsolatedAsyncioTestCase):

    &#34;&#34;&#34;
    获取的是整个网页的完整 HTML 内容
    &#34;&#34;&#34;
    async def test_playwright(self):
        async with async_playwright() as p:
            browser = await p.chromium.connect(&#34;ws://ip:3000/&#34;)
            page = await browser.new_page()
            await page.goto(&#34;https://example.com&#34;)
            content = await page.content()
            print(content)
            await browser.close()

    &#34;&#34;&#34;
    提取文本
    &#34;&#34;&#34;
    async def test_playwright_extract_txt(self):
        async with async_playwright() as p:
            browser = await p.chromium.connect(&#34;ws://ip:3000/&#34;)
            page = await browser.new_page()
            await page.goto(&#34;https://example.com&#34;)

            # 只要纯文本，不要 HTML 标签
            # 使用 JavaScript 提取纯文本
            text_content = await page.evaluate(&#39;&#39;&#39;() =&gt; {
                return document.body.innerText || document.documentElement.innerText;
            }&#39;&#39;&#39;)

            # 清洗格式（空行、缩进等）
            # text_content = await page.evaluate(&#39;&#39;&#39;() =&gt; {
            #     return document.body.innerText.replace(/\\s&#43;/g, &#39; &#39;).trim();
            # }&#39;&#39;&#39;)

            print(text_content)

            await browser.close()


if __name__ == &#34;__main__&#34;:
    unittest.main()
```



### 设置请求头（Headers）

``` python {data-open=true,title=&#34;设置请求头（Headers）&#34;}
async def test_playwright():
    async with async_playwright() as p:
        # 连接远程浏览器
        browser = await p.chromium.connect(&#34;ws://ip:3000/&#34;)

        # 创建带有自定义 headers 的上下文
        context = await browser.new_context(extra_http_headers={
            &#34;User-Agent&#34;: &#34;MyCustomUserAgent/1.0&#34;,
            &#34;Authorization&#34;: &#34;Bearer YOUR_TOKEN_HERE&#34;,
            &#34;X-Custom-Header&#34;: &#34;SomeValue&#34;
        })

        # 在该上下文中打开新页面
        page = await context.new_page()
        await page.goto(&#34;https://example.com&#34;)

        # 获取内容或其他操作
        content = await page.content()
        print(content)

        # 关闭 context 和 browser（注意：不要关闭远程浏览器实例）
        await context.close()
        # await browser.close()  # 不建议关闭远程浏览器连接，除非你知道自己在做什么
```


### 抓取多个网页

``` python {data-open=true,title=&#34;抓取多个网页&#34;}
import asyncio
from playwright.async_api import async_playwright

async def test_playwright():
    urls = [
        &#34;https://example.com&#34;,
        &#34;https://example.org&#34;,
        &#34;https://example.net&#34;
    ]
    
    async with async_playwright() as p:
        try:
            # 连接远程浏览器
            browser = await p.chromium.connect(&#34;ws://ip:3000/&#34;)
            
            # 创建带有自定义 headers 的上下文
            context = await browser.new_context(extra_http_headers={
                &#34;User-Agent&#34;: &#34;MyCustomUserAgent/1.0&#34;,
                &#34;Authorization&#34;: &#34;Bearer YOUR_TOKEN_HERE&#34;,
                &#34;X-Custom-Header&#34;: &#34;SomeValue&#34;
            })

            for url in urls:
                try:
                    # 在该上下文中打开新页面
                    page = await context.new_page()
                    await page.goto(url)
                    
                    # 获取内容或其他操作
                    content = await page.content()
                    print(f&#34;Content of {url}:&#34;)
                    print(content)
                
                except Exception as e:
                    print(f&#34;Error while processing {url}: {e}&#34;)
                
                finally:
                    # 关闭当前页面
                    if &#39;page&#39; in locals():
                        await page.close()

        except Exception as e:
            print(f&#34;An error occurred: {e}&#34;)

        finally:
            # 确保上下文被关闭
            if &#39;context&#39; in locals():
                await context.close()

# 运行异步函数
asyncio.run(test_playwright())
```


### 整体例子

``` python {data-open=true,title=&#34;整体例子&#34;}
&#34;&#34;&#34;
抓取url网址内容
&#34;&#34;&#34;
async def fetch_page(doc:Dict[str,Any], currLevel: int = 1, maxLevel: int = 1, kbId: str = None, white_list: List[str] = None, loadURL:Dict[str,int]={}, expression:List[str] = None, headers: Dict[str, Any] ={}):
    url = doc.get(&#34;path&#34;)
    if currLevel &gt; maxLevel:
        return
    if loadURL.get(url) is not None:
        return
    loadURL[url] = 1
    # 检查是否有文件被上传
    # 使用 Playwright 抓取网页
    async with async_playwright() as p:
        try:
            browser = await p.chromium.connect(yamlConfig.get(&#34;playwright&#34;).get(&#34;url&#34;))

            # 创建带有自定义 headers 的上下文
            context = await browser.new_context(extra_http_headers=headers)

            # 在该上下文中打开新页面
            page = await context.new_page()
            # networkidle:至少500毫秒没有网络连接活动时，Playwright 将认为页面已加载完毕
            await page.goto(url, wait_until=&#39;networkidle&#39;)

            raw_html = await page.content() # HTML 内容(主要用于获取页面标题)
            links = []      # 页面中的链接
            full_text = &#34;&#34;  # 纯文本

            # 如果有XPath表达式
            if len(expression) &gt; 0 :
                # 处理每个 XPath 表达式
                for xpath in expression:
                    elements_html = await get_elements_by_xpath(page, xpath)
                    if not elements_html:
                        continue

                    # 每个 XPath 可能匹配多个元素
                    for html in elements_html:
                        soup = BeautifulSoup(html, &#39;lxml&#39;)
                        # 提取链接并加入到全局links列表中
                        links.extend([a.get(&#39;href&#39;) for a in soup.find_all(&#39;a&#39;, href=True) if a.get(&#39;href&#39;) not in links])
                        # 提取纯文本并拼接到full_text
                        full_text &#43;= soup.get_text(separator=&#39; &#39;, strip=True) &#43; &#34; &#34;

            else:
                # 使用 BeautifulSoup 解析 HTML
                soup = BeautifulSoup(raw_html, &#34;lxml&#34;)

                # 提取页面中的所有链接
                links = {a.get(&#39;href&#39;) for a in soup.find_all(&#39;a&#39;, href=True)}

                # 纯文本内容
                full_text = await page.evaluate(&#39;&#39;&#39;() =&gt; {
                    return document.body.innerText.replace(/\\s&#43;/g, &#39; &#39;).trim();
                }&#39;&#39;&#39;)

            # 关闭页面
            await page.close()

            if len(full_text) == 0:
                return ReturnDatas.ErrorResponse(message=&#34;The uploaded webpage text is empty!&#34;)

            # 查找 &lt;title&gt; 标签并获取其内容
            start_title = raw_html.find(&#39;&lt;title&gt;&#39;)
            end_title = raw_html.find(&#39;&lt;/title&gt;&#39;)

            if start_title != -1 and end_title != -1:
                # 提取 &lt;title&gt; 中间的文本内容
                title = raw_html[start_title &#43; 7:end_title].strip()
            else:
                title = &#34;No Title Found&#34;

            if len(title) == 0:
                return ReturnDatas.ErrorResponse(message=&#34;The title is empty!&#34;)

        except Exception as e:
            core_logger.exception(e)
            traceback.print_exc()


        finally:
            if &#39;context&#39; in locals():
                await context.close()


&#34;&#34;&#34;
根据 XPath 表达式获取标签
&#34;&#34;&#34;
async def get_elements_by_xpath(page, xpath_expr):
    # 使用 page.locator() 方法结合 XPath 选择器
    locator = page.locator(f&#39;xpath={xpath_expr}&#39;)

    # 获取所有匹配的元素数量
    count = await locator.count()
    if count == 0:
        return []

    results = []
    for i in range(count):
        # 获取每个元素的 HTML 内容
        html = await locator.nth(i).evaluate(&#39;e =&gt; e.outerHTML&#39;)
        results.append(html)

    return results
```











---
{.awesome-hr}

---

> 作者: [piliqiu](https://piliqiu.com/)  
> URL: https://piliqiu.com/posts/83e1aa3/  

